# tests.py
"""Defines tests for :class:`Trie`, :class:`Parser`, :func:`parser_factory`.

"""


import unittest
import parser
from parser import *
import trie


class TestTrie(unittest.TestCase):

    def test_one_entry(self):
        t = trie.Trie()
        t.add_string('ABC')
        self.assertTrue('ABC' in t)
        self.assertTrue(t.root == {'A' : {'B': {'C': None}}})

    def test_two_entry(self):
        t = trie.Trie()
        t.add_string('ABC')
        t.add_string('Axx')
        self.assertTrue('ABC' in t)

    def test_same_first_middle(self):
        t = trie.Trie()
        t.add_string('ABC')
        t.add_string('ABX')
        self.assertTrue('ABC' in t)
        self.assertTrue('ABX' in t)

    def test_different_first(self):
        t = trie.Trie()
        t.add_string('ABC')
        t.add_string('ZBC')
        self.assertTrue('ABC' in t)
        self.assertTrue('ZBC' in t)

    def test_no_match_first(self):
        t = trie.Trie()
        t.add_string('ABC')
        self.assertFalse('?BC' in t)

    def test_no_match_mid(self):
        t = trie.Trie()
        t.add_string('ABC')
        self.assertFalse('A?C' in t)

    def test_no_match_last(self):
        t = trie.Trie()
        t.add_string('ABC')
        self.assertFalse('AB?' in t)

    def test_different_end(self):
        t = trie.Trie()
        t.add_string('ABC')
        t.add_string('ABX')
        self.assertFalse('AB?' in t)


class TestParserFactory(unittest.TestCase):


    def test_chinese_simplified(self):
        zh_Hans = parser_factory('zh-Hans')
        zh_Hant = parser_factory('zh-Hant')
        yue_Hans = parser_factory('yue-Hans')
        yue_Hant = parser_factory('yue-Hant')
        self.assertTrue(type(zh_Hans) is ChineseSimplifiedParser)
        self.assertTrue(type(zh_Hant) is ChineseTraditionalParser)
        self.assertTrue(type(yue_Hans) is CantoneseSimplifiedParser)
        self.assertTrue(type(yue_Hant) is CantoneseTraditionalParser)

class TestParserBaseClass(object):
    """Abstract base class for testing parsers."""

    def setUp(self):
        self.p = self.parser()
    
    def test_multiple_tokens(self):
        string = ''.join(x for x in self.tokens)
        result = self.p.parse(string)
        self.assertEqual(set(result), self.tokens)

    def test_empty_str(self):
        self.assertEqual(self.p.parse(''), [])


class TestChineseSimplifiedParser(unittest.TestCase):


    def setUp(self):
        self.p = ChineseSimplifiedParser()

    def test_single_token(self):
        result = self.p.parse('感')
        self.assertEqual(['感'], result)


    def test_first_string_match(self):
        result = self.p.parse('感?')
        self.assertEqual(['感','?'], result)

    def test_two_word_sequence(self):
       result = self.p.parse('政府感同身受')
       self.assertEqual(['政府', '感同身受'], result)

    def test_two_similar_words(self):
       result = self.p.parse('非政府政府')
       self.assertEqual(['非政府', '政府'], result)

    def test_matches_and_nonmatches_mixed(self):
       result = self.p.parse('受非政府盈政府')
       self.assertEqual(['受','非政府', '盈','政府'], result)

    def test_random(self):
        self.assertTrue('感' in self.p.trie)


class TestChineseParser(TestParserBaseClass, unittest.TestCase):
    """Define unique tokens in a set, the parser, and the TestParserBaseClass will automatically test
    the parser's functionality.  Tests that ChineseParser parses simplfiied
    and traditional characters.
    """

    tokens = {'鳳凰古城', '鮮明', '鬧翻', '非政府','感同身受'}
    parser = ChineseParser



class TestChineseTraditionalParser(TestParserBaseClass, unittest.TestCase):
    """Define unique tokens in a set, the parser, and the TestParserBaseClass will automatically test
    the parser's functionality.
    """

    tokens = {'鳳凰古城', '鮮明', '鬧翻'}
    parser = ChineseTraditionalParser


class TestCantoneseParser(TestParserBaseClass, unittest.TestCase):
    tokens = {'三八线', '不避艰险', '中国光大银行', '不計其數', '二鬼子', '人頭'}
    parser = CantoneseParser


class TestCantoneseSimplifiedParser(TestParserBaseClass, unittest.TestCase):
    tokens = {'三八线', '不避艰险', '中国光大银行'}
    parser = CantoneseSimplifiedParser


class TestCantoneseTraditionalParser(TestParserBaseClass, unittest.TestCase):
    tokens = {'不計其數', '二鬼子', '人頭'}
    parser = CantoneseTraditionalParser


class TestThaiParser(TestParserBaseClass, unittest.TestCase):
    tokens = {'รูปสี่เหลี่ยมขนมเปียกปูน', 'ทิ้งลูกทิ้งเมีย', 'ผู้ที่สูงอายุ'}
    parser = ThaiParser


class TestParser(unittest.TestCase):


    def test_reference_error(self):
        with self.assertRaises(FileNotFoundError):
            temp = Parser('garbage/path')
    

if __name__ == '__main__':
    unittest.main()



import collections.abc
from itertools import zip_longest


class Trie(collections.abc.Container):

    def __init__(self):
        self.root = dict()

    def add_string(self, string):
        current_node = self.root
        for pair in zip_longest(string, string[1:]):
            if (pair[0] not in current_node and
                pair[1] is not None):
                current_node[pair[0]] = { pair[1] : dict() }
            elif (pair[0] not in current_node and
                pair[1] is None):
                current_node[pair[0]] = pair[1]
            current_node = current_node[pair[0]]

    def __contains__(self, seq):
        current_node = self.root
        try:
            for elem in seq:
                current_node = current_node[elem]
        except KeyError:
            return False
        return True

    def find_match(self, string):
        current_node = self.root
        try:
            for c in string:
                current_node = current_node[c]
        except KeyError:
            return False
        return True



# parser.py
"""This module provides classes for parsing strings of natural languages
that do not use space delimiters.


Currently Supported Languages
=============================

====================== ============= ==========================
Language               Language Code Parser Class
====================== ============= ==========================
Mandarin               zh            ChineseParser
Mandarin Simplified    zh-Hans       ChineseSimplifiedParser
Mandarin Traditional   zh-Hant       ChineseTraditionalParser
Cantonese              yue           CantoneseParser
Cantonese Simplified   yue-Hans      CantoneseSimplifiedParser
Cantonese Traditional  yue-Hant      CantoneseTraditionalParser
Thai                   th or thai    ThaiParser
====================== ============= ==========================


"""

import os
import pickle
from wordfencer.trie import Trie
from wordfencer.exceptions import ParserError


def parser_factory(lang):
    """Returns a Parser object specific to the language specified by :param:`lang`.

    :param lang: A string naming the desired parser.  See the keys in :var:`class_`
    for valid strings.

    :Example:

    >>> chinese_parser = parser_factory('zh-Hans')
    >>> isinstance(chinese_parser, ChineseSimplifiedParser)
    True

    >>> cantonese_parser = parser_factory('yue-Hant')
    >>> isinstance(cantonese_parser, CantoneseTraditionalParser)
    True

    """

    class_ = {
            'zh-Hans' : ChineseSimplifiedParser,
            'zh-Hant' : ChineseTraditionalParser,
            'zh' : ChineseParser,
            'yue-Hans' : CantoneseSimplifiedParser,
            'yue-Hant' : CantoneseTraditionalParser,
            'yue' : CantoneseParser,
            'thai' : ThaiParser,
            'th' : ThaiParser,
            }.get(lang)
    return class_()


class Parser(object):

    """Defines a generic class for parsing languages with no space delimiters.

    `Parser` uses a pre-existing reference dictionary to build a Trie.
    :func:`force_populate` builds the trie out of the reference dictionary file.
    :func:`parse` returns a list of words parsed from :arg:`string`.  The longest
    segment is always selected.


    Attributes
    ==========

    trie : Trie data structure
        The trie used to perform parsing algorithm.
    populated : boolean
        True if the trie has already been populated with words, otherwise False.
    ref : string
        The relative path to the reference dictionary file.
    file : string
        The relative path to the pickle file holding a previously built trie.


    """

    def __init__(self, name):
        self.ref = os.path.dirname(__file__) + '/data/' + name + '.txt'
        self.file = os.path.dirname(__file__) + '/data/' + name + '.pickle'
        self.trie = Trie()
        self.populated = False
        self.load()

    def parse(self, string, all_combos=False):
        """Returns a list of words created from segmenting :arg:`string`.

        ..note:  When a series of characters could match a shorter dictionary entry or
        longer entry, the longer one is always selected.
        """
        if not all_combos:
            return self.default_parse(string)
        else:
            return self.all_combinations(string)


    def default_parse(self, string):
        """Returns a list of words created from segmenting :arg:`string`.

        ..note:  When a series of characters could match a shorter dictionary entry or
        longer entry, the longer one is always selected.
        """
        if not self.populated:
            raise ParserError('Parser not yet populated, must call force_populate()')
        results = list()
        while(string):
            token = self.__next_token(string)
            if token:
                results.append(token)
                string = string[len(token):]
            else:
                string = ''
        return results

    def all_combinations(self, string):
        token_set = set()
        for i in range(len(string)):
            for elem in self.default_parse(string[i:]):
                token_set.add(elem)
        return token_set

    def __next_token(self, string):
        """Returns the next token from the :arg:`string`.
        """
        temp = ''
        for c in string:
            temp += c
            if temp not in self.trie:
                if len(temp) > 1:
                    return temp[:-1]
                if len(temp) == 1:
                    return temp
        return temp

    def load(self):
        try:
            with open(self.file, 'rb') as f:
                self.trie = pickle.load(f)
                self.populated = True
        except FileNotFoundError:
            self.force_populate()
            self.save()

    def force_populate(self):
        """
        Populates the parser with the entire contents of the 
        word reference file.
        """
        if not os.path.exists(self.ref):
            raise FileNotFoundError("The reference file path '{}' does not exists.".format(self.ref))
        with open(self.ref, 'r') as f:
            for word in f:
                self.trie.add_string(word)
        self.populated = True

    def save(self):
        with open(self.file, 'wb') as f:
            pickle.dump(self.trie, f)


class ChineseParser(Parser):
    """This parser is for Mandarin Chinese written with Simplified OR Traditional characters.

    """

    def __init__(self):
        super().__init__('zh')


class ChineseSimplifiedParser(Parser):
    """This parser is for Mandarin Chinese written with Simplified characters.

    """

    def __init__(self):
        super().__init__('zh-Hans')


class ChineseTraditionalParser(Parser):
    """This parser is for Mandarin Chinese written with Traditional characters.

    All attributes and functionality inherited from :class:`Parser`.

    """

    def __init__(self):
        super().__init__('zh-Hant')


class CantoneseParser(Parser):
    """This parser is for Cantonese written with Simplified OR Traditional characters.

    """

    def __init__(self):
        super().__init__('yue')


class CantoneseSimplifiedParser(Parser):
    """This parser is for Cantonese written with Simplified characters.

    All attributes and functionality inherited from :class:`Parser`.


    """

    def __init__(self):
        super().__init__('yue-Hans')


class CantoneseTraditionalParser(Parser):
    """This parser is for Cantonese written with Traditional characters.

    All attributes and functionality inherited from :class:`Parser`.


    """

    def __init__(self):
        super().__init__('yue-Hant')


class ThaiParser(Parser):
    """This parser is for Thai.

    All attributes and functionality inherited from :class:`Parser`.


    """

    def __init__(self):
        super().__init__('thai')


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# setup.py
"""This module simply reads a dictionary reference file, builds a Trie, and then pickles it.
"""

from setuptools import setup, find_packages
setup(
    name = "wordfencer",
    packages = find_packages(),
    version = "1.1.5",
    description = "Parser for natural languages without space delimiters",
    author = "Bobby Eshleman",
    author_email = "bobbyeshleman@gmail.com",
    url = "https://github.com/bobbyesh/word-fencer",
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Natural Language :: Cantonese",
        "Natural Language :: Chinese (Simplified)",
        "Natural Language :: Chinese (Traditional)",
        "Natural Language :: Thai",
    ],
    include_package_data = True,
    package_data = { '' : ['*.pickle'] },
    long_description = """\
Parser For Chinese, Cantonese, and Thai
---------------------------------------

Tokenizes the words found in any arbitrary string in any of the supported languages.
""",
)
